package com.sun.corba.se.impl.encoding;

import com.sun.corba.se.impl.orbutil.ORBUtility;
import java.util.ListIterator;
import com.sun.corba.se.impl.protocol.giopmsgheaders.Message;
import com.sun.corba.se.pept.transport.ByteBufferPool;
import com.sun.corba.se.impl.protocol.RequestCanceledException;
import com.sun.corba.se.impl.protocol.giopmsgheaders.FragmentMessage;
import java.nio.ByteBuffer;
import java.util.LinkedList;
import com.sun.corba.se.impl.logging.ORBUtilSystemException;
import com.sun.corba.se.spi.orb.ORB;

public class BufferManagerReadStream implements BufferManagerRead, MarkAndResetHandler
{
    private boolean receivedCancel;
    private int cancelReqId;
    private boolean endOfStream;
    private BufferQueue fragmentQueue;
    private long FRAGMENT_TIMEOUT;
    private ORB orb;
    private ORBUtilSystemException wrapper;
    private boolean debug;
    private boolean markEngaged;
    private LinkedList fragmentStack;
    private RestorableInputStream inputStream;
    private Object streamMemento;
    
    BufferManagerReadStream(ORB orb) {
        this.receivedCancel = false;
        this.cancelReqId = 0;
        this.endOfStream = true;
        this.fragmentQueue = new BufferQueue();
        this.FRAGMENT_TIMEOUT = 60000L;
        this.debug = false;
        this.markEngaged = false;
        this.fragmentStack = null;
        this.inputStream = null;
        this.streamMemento = null;
        this.orb = orb;
        this.wrapper = ORBUtilSystemException.get(orb, "rpc.encoding");
        this.debug = orb.transportDebugFlag;
    }
    
    @Override
    public void cancelProcessing(int cancelReqId) {
        synchronized (this.fragmentQueue) {
            this.receivedCancel = true;
            this.cancelReqId = cancelReqId;
            this.fragmentQueue.notify();
        }
    }
    
    @Override
    public void processFragment(ByteBuffer byteBuffer, FragmentMessage fragmentMessage) {
        ByteBufferWithInfo byteBufferWithInfo = new ByteBufferWithInfo(this.orb, byteBuffer, fragmentMessage.getHeaderLength());
        synchronized (this.fragmentQueue) {
            if (this.debug) {
                int identityHashCode = System.identityHashCode(byteBuffer);
                StringBuffer sb = new StringBuffer(80);
                sb.append("processFragment() - queueing ByteBuffer id (");
                sb.append(identityHashCode).append(") to fragment queue.");
                this.dprint(sb.toString());
            }
            this.fragmentQueue.enqueue(byteBufferWithInfo);
            this.endOfStream = !fragmentMessage.moreFragmentsToFollow();
            this.fragmentQueue.notify();
        }
    }
    
    @Override
    public ByteBufferWithInfo underflow(ByteBufferWithInfo byteBufferWithInfo) {
        ByteBufferWithInfo dequeue = null;
        synchronized (this.fragmentQueue) {
            if (this.receivedCancel) {
                throw new RequestCanceledException(this.cancelReqId);
            }
            while (this.fragmentQueue.size() == 0) {
                if (this.endOfStream) {
                    throw this.wrapper.endOfStream();
                }
                boolean n = false;
                try {
                    this.fragmentQueue.wait(this.FRAGMENT_TIMEOUT);
                }
                catch (InterruptedException ex) {
                    n = true;
                }
                if (n == 0 && this.fragmentQueue.size() == 0) {
                    throw this.wrapper.bufferReadManagerTimeout();
                }
                if (this.receivedCancel) {
                    throw new RequestCanceledException(this.cancelReqId);
                }
                continue;
            }
            dequeue = this.fragmentQueue.dequeue();
            dequeue.fragmented = true;
            if (this.debug) {
                int identityHashCode = System.identityHashCode(dequeue.byteBuffer);
                StringBuffer o = new StringBuffer(80);
                o.append("underflow() - dequeued ByteBuffer id (");
                o.append(identityHashCode).append(") from fragment queue.");
                this.dprint(o.toString());
            }
            if (!this.markEngaged && byteBufferWithInfo != null && byteBufferWithInfo.byteBuffer != null) {
                ByteBufferPool byteBufferPool = this.getByteBufferPool();
                if (this.debug) {
                    int identityHashCode2 = System.identityHashCode(byteBufferWithInfo.byteBuffer);
                    StringBuffer o2 = new StringBuffer(80);
                    o2.append("underflow() - releasing ByteBuffer id (");
                    o2.append(identityHashCode2).append(") to ByteBufferPool.");
                    this.dprint(o2.toString());
                }
                byteBufferPool.releaseByteBuffer(byteBufferWithInfo.byteBuffer);
                byteBufferWithInfo.byteBuffer = null;
            }
        }
        return dequeue;
    }
    
    @Override
    public void init(Message message) {
        if (message != null) {
            this.endOfStream = !message.moreFragmentsToFollow();
        }
    }
    
    @Override
    public void close(ByteBufferWithInfo byteBufferWithInfo) {
        int n = 0;
        if (this.fragmentQueue != null) {
            synchronized (this.fragmentQueue) {
                if (byteBufferWithInfo != null) {
                    n = System.identityHashCode(byteBufferWithInfo.byteBuffer);
                }
                ByteBufferPool byteBufferPool = this.getByteBufferPool();
                while (this.fragmentQueue.size() != 0) {
                    ByteBufferWithInfo dequeue = this.fragmentQueue.dequeue();
                    if (dequeue == null || dequeue.byteBuffer == null) {
                        continue;
                    }
                    int identityHashCode = System.identityHashCode(dequeue.byteBuffer);
                    if (n != identityHashCode && this.debug) {
                        StringBuffer o = new StringBuffer(80);
                        o.append("close() - fragmentQueue is ").append("releasing ByteBuffer id (").append(identityHashCode).append(") to ").append("ByteBufferPool.");
                        this.dprint(o.toString());
                    }
                    byteBufferPool.releaseByteBuffer(dequeue.byteBuffer);
                }
            }
            this.fragmentQueue = null;
        }
        if (this.fragmentStack != null && this.fragmentStack.size() != 0) {
            if (byteBufferWithInfo != null) {
                n = System.identityHashCode(byteBufferWithInfo.byteBuffer);
            }
            ByteBufferPool byteBufferPool2 = this.getByteBufferPool();
            ListIterator<Object> listIterator = this.fragmentStack.listIterator();
            while (listIterator.hasNext()) {
                ByteBufferWithInfo byteBufferWithInfo2 = (ByteBufferWithInfo)listIterator.next();
                if (byteBufferWithInfo2 == null || byteBufferWithInfo2.byteBuffer == null) {
                    continue;
                }
                int identityHashCode2 = System.identityHashCode(byteBufferWithInfo2.byteBuffer);
                if (n != identityHashCode2) {
                    if (this.debug) {
                        StringBuffer o2 = new StringBuffer(80);
                        o2.append("close() - fragmentStack - releasing ").append("ByteBuffer id (" + identityHashCode2 + ") to ").append("ByteBufferPool.");
                        this.dprint(o2.toString());
                    }
                    byteBufferPool2.releaseByteBuffer(byteBufferWithInfo2.byteBuffer);
                }
                else {
                    continue;
                }
            }
            this.fragmentStack = null;
        }
    }
    
    protected ByteBufferPool getByteBufferPool() {
        return this.orb.getByteBufferPool();
    }
    
    private void dprint(String s) {
        ORBUtility.dprint("BufferManagerReadStream", s);
    }
    
    @Override
    public void mark(RestorableInputStream inputStream) {
        this.inputStream = inputStream;
        this.markEngaged = true;
        this.streamMemento = inputStream.createStreamMemento();
        if (this.fragmentStack != null) {
            this.fragmentStack.clear();
        }
    }
    
    @Override
    public void fragmentationOccured(ByteBufferWithInfo byteBufferWithInfo) {
        if (!this.markEngaged) {
            return;
        }
        if (this.fragmentStack == null) {
            this.fragmentStack = new LinkedList();
        }
        this.fragmentStack.addFirst(new ByteBufferWithInfo(byteBufferWithInfo));
        return;
    }
    
    @Override
    public void reset() {
        if (!this.markEngaged) {
            return;
        }
        this.markEngaged = false;
        if (this.fragmentStack != null && this.fragmentStack.size() != 0) {
            ListIterator<Object> listIterator = this.fragmentStack.listIterator();
            synchronized (this.fragmentQueue) {
                while (listIterator.hasNext()) {
                    this.fragmentQueue.push((ByteBufferWithInfo)listIterator.next());
                }
            }
            this.fragmentStack.clear();
        }
        this.inputStream.restoreInternalState(this.streamMemento);
        return;
    }
    
    @Override
    public MarkAndResetHandler getMarkAndResetHandler() {
        return this;
    }
}
